@@ -7,7 +7,6 @@
 import atexit
 import logging
 import os
-import time
 
 import attr
 
@@ -18,6 +17,7 @@ from saltfactories.factories.base import SaltDaemonFactory
 from saltfactories.factories.daemons.minion import SaltMinionFactory
 from saltfactories.utils import ports
 from saltfactories.utils import random_string
+from saltfactories.utils import time
 from saltfactories.utils.processes import ProcessResult
 
 try:
@@ -86,7 +86,7 @@ class ContainerFactory(Factory):
     def _format_callback(self, callback, args, kwargs):
         callback_str = "{}(".format(callback.__name__)
         if args:
-            callback_str += ", ".join(args)
+            callback_str += ", ".join([repr(arg) for arg in args])
         if kwargs:
             callback_str += ", ".join(["{}={!r}".format(k, v) for (k, v) in kwargs.items()])
         callback_str += ")"
